<!-- default banner, create through js -->
<script src='{{ wbrequest.host_prefix }}/{{ static_path }}/default_banner.js'> </script>
<link rel='stylesheet' href='{{ wbrequest.host_prefix }}/{{ static_path }}/wb.css'/>

