<!DOCTYPE html>
<html>
<head>
<style>
html, body
{
  height: 100%;
  margin: 0px;
  padding: 0px;
  border: 0px;
  overflow: hidden;
}

</style>
<script>
  wbinfo = {}
  wbinfo.prefix = "{{ wbrequest.wb_prefix }}";
  wbinfo.capture_url = "{{ url }}";
  wbinfo.timestamp = "{{ wbrequest.wb_url.timestamp }}";
  wbinfo.is_frame = true;
  wbinfo.frame_mod = "{{ wbrequest.options.frame_mod }}";
  wbinfo.replay_mod = "{{ wbrequest.options.replay_mod }}";
</script>
<script src='{{ wbrequest.host_prefix }}/{{ static_path }}/wb.js'> </script>
<link rel='stylesheet' href='{{ wbrequest.host_prefix }}/{{ static_path }}/wb.css'/>

{% include banner_html ignore missing %}

<script src='{{ wbrequest.host_prefix }}/{{ static_path }}/wb_frame.js'> </script>
</head>
<body style="margin: 0px; padding: 0px;">
<div class="wb_iframe_div">
<iframe id="replay_iframe" src="{{ wbrequest.wb_prefix + embed_url }}" onload="iframe_loaded(event);" seamless="seamless" frameborder="0" scrolling="yes" class="wb_iframe"></iframe>
</div>
</body>
</html>
